@@ -32,7 +32,7 @@ import asyncio
 import shutil
 import websockets
 from uuid import uuid4
-from bokeh.models import Button, TextInput, Div, Range1d, LinearAxis, CustomJS
+from bokeh.models import Button, TextInput, Div, Range1d, LinearAxis, CustomJS, Spacer
 from bokeh.plotting import ColumnDataSource, figure, show
 from bokeh.layouts import column, row, Spacer
 from ..utils import resource_manager
@@ -87,7 +87,7 @@ class InteractiveClean:
     of these boxes are updated as the clean operation progresses.
 
     The cleaning interval is determined by the text inputs along the top of the GUI which
-    allow the user to set the ``niter``, ``cycleiter``, ``cyclefactor``, and ``threshold``.
+    allow the user to set the ``nmajor``, ``niter``, ``cycleiter``, ``cyclefactor``, and ``threshold``.
     The three buttons ( _run one interval_, _run until stoping criteral is reached_, and
     _stop_) allow the user to control the clean run.
 
@@ -123,6 +123,7 @@ class InteractiveClean:
         gridder (str): Gridding options (standard, wproject, widefield, mosaic, awproject)
         pblimit (float): PB gain level at which to cut off normalizations
         deconvolver (str): Minor cycle algorithm (hogbom,clark,multiscale,mtmfs,mem,clarkstokes)
+        nmajor (int): Maximum number of major cycle iterations
         niter (int): Maximum total number of iterations
         threshold (str | float): Stopping threshold (number in units of Jy, or string).
         cycleniter (int): Maximum number of minor-cycle iterations (per plane) before triggering a major cycle
@@ -141,14 +142,15 @@ class InteractiveClean:
         self._error_result = err
         self.__stop( )
 
-    def __init__( self, vis, imagename, imsize=[100], cell="1arcsec", specmode='cube', nchan=-1, start='',
-                  width='', interpolation='linear', gridder='standard', pblimit=0.2, deconvolver='hogbom',
-                  niter=0, threshold='0.1Jy', cycleniter=-1, cyclefactor=1.0, scales=[] ):
+    def __init__( self, vis, imagename, field='', spw='', imsize=[100], cell=[ ], phasecenter='', stokes='I', specmode='cube', nchan=-1, start='',
+                  width='', interpolation='linear', gridder='standard', pblimit=0.2, deconvolver='hogbom', niter=0,
+                  threshold='0.1Jy', cycleniter=-1, cyclefactor=1.0, scales=[], weighting='natural', robust=float(0.5), nmajor=1 ):
 
         ###
         ### used by data pipe (websocket) initialization function
         ###
         self.__pipes_initialized = False
+        self._mask_history = [ ]
 
         ###
         ### color specs
@@ -162,10 +164,10 @@ class InteractiveClean:
         if _gclean is None:
             raise RuntimeError('casatasks gclean interface is not available')
 
-        self._clean = _gclean( vis=vis, imagename=imagename, imsize=imsize, cell=cell, specmode=specmode, nchan=nchan,
-                               start=start, width=width, interpolation=interpolation, gridder=gridder, pblimit=pblimit,
-                               deconvolver=deconvolver, niter=niter, threshold=threshold, cycleniter=cycleniter,
-                               cyclefactor=cyclefactor, scales=scales,
+        self._clean = _gclean( vis=vis, imagename=imagename, field=field, spw=spw, imsize=imsize, cell=cell, phasecenter=phasecenter,
+                               stokes=stokes, specmode=specmode, nchan=nchan, start=start, width=width, interpolation=interpolation,
+                               gridder=gridder, pblimit=pblimit, deconvolver=deconvolver, niter=niter, threshold=threshold,
+                               cycleniter=cycleniter, cyclefactor=cyclefactor, scales=scales, weighting=weighting, robust=robust, nmajor=nmajor,
                                history_filter=lambda index, arg, history_value: ( f'mask=masks[{len(self._mask_history)-1}]' if len(self._mask_history) > 0 else '' ) \
                                                                                   if arg == 'mask' else history_value )
         ###
@@ -188,6 +190,7 @@ class InteractiveClean:
         ### Initial Conditions
         ###
         self._params = { }
+        self._params['nmajor'] = nmajor
         self._params['niter'] = niter
         self._params['cycleniter'] = cycleniter
         self._params['threshold'] = threshold
@@ -206,7 +209,6 @@ class InteractiveClean:
         self._cb = { }
         self._ids = { }
         self._last_mask_breadcrumbs = ''
-        self._mask_history = [ ]
 
         ###
         ### The tclean convergence data is automatically generated by tclean and it
@@ -256,7 +258,12 @@ class InteractiveClean:
                                                //                              stat_src.data = msg.stats
                                                //                          }
                                                //                        } )
-                                               img_src.refresh( )
+                                               if ( clean_msg !== undefined && 'iterdone' in clean_msg ) {
+                                                 const remaining = parseInt(niter.value) - parseInt(clean_msg['iterdone'])
+                                                 niter.value = '' + (remaining < 0 ? 0 : remaining)
+                                               }
+                                               img_src.refresh( (data) => { if ( 'stats' in data ) cube_obj.update_statistics( data.stats ) } )
+
                                                if ( clean_msg !== undefined && 'convergence' in clean_msg ) {
                                                    // save convergence information and update convergence using saved state
                                                    if ( clean_msg.convergence === null ) {
@@ -279,6 +286,7 @@ class InteractiveClean:
                                            // self._fig['image'].toolbar.tools.tool_name (e.g. "Box Select", "Lasso Select")
                                            function disable( with_stop ) {
                                                img_src.disable_masking( )
+                                               nmajor.disabled = true
                                                niter.disabled = true
                                                cycleniter.disabled = true
                                                threshold.disabled = true
@@ -297,11 +305,11 @@ class InteractiveClean:
 
                        'clean-enable':  '''function enable( only_stop ) {
                                                img_src.enable_masking( )
+                                               nmajor.disabled = false
                                                niter.disabled = false
                                                cycleniter.disabled = false
                                                threshold.disabled = false
                                                cyclefactor.disabled = false
-                                               niter.disabled = false
                                                btns['stop'].disabled = false
                                                slider.disabled = false
                                                image_fig.disabled = false
@@ -368,7 +376,7 @@ class InteractiveClean:
                                                        if ( ! state.stopped ) {
                                                            ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                            { action: 'finish',
-                                                                             value: { niter: niter.value, cycleniter: cycleniter.value,
+                                                                             value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                                       threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                                       mask: img_src.masks( ),
                                                                                       breadcrumbs: img_src.breadcrumbs( ) } },
@@ -444,29 +452,31 @@ class InteractiveClean:
                         shutil.rmtree(mask_dir)
                         new_mask = self._cube.jsmask_to_raw(msg['value']['mask'])
                         self._mask_history.append(new_mask)
-                        ##msg['value']['mask'] = convert_masks( new_mask, 'region','pixel','singleton', cdesc=self._cube.coorddesc( ) )
-                        msg['value']['mask'] = convert_masks( new_mask, 'crtf', 'pixel', 'list' )
+
+                        msg['value']['mask'] = convert_masks(masks=new_mask, coord='pixel', cdesc=self._cube.coorddesc())
+
                     else:
                         msg['value']['mask'] = ''
                 else:
                     msg['value']['mask'] = ''
-                self._clean.update( { **msg['value'], 'nmajor': 1 if msg['action'] == 'next' else -1 } )
+                self._clean.update( { **msg['value'] } )
                 stopcode, self._convergence_data = await self._clean.__anext__( )
                 if stopcode > 1 and stopcode < 9: # 1: iteration limit hit, 9: major cycle limit hit
                     self._clean.finalize()
+
                     # self._cube.update_image(self._clean.finalize()['image']) # TODO show the restored image
                 if len(self._convergence_data) == 0 and stopcode == 7:
                     return dict( result='error', stopcode=stopcode, cmd=f"<p>mask error encountered (stopcode {stopcode})</p>", convergence=None  )
                 if len(self._convergence_data) * len(self._convergence_data[0]) > self._threshold_chan or \
                    len(self._convergence_data[0][0]['iterations']) > self._threshold_iterations:
                     return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
-                                 convergence=None )
+                                 convergence=None, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
                 else:
                     return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
-                                 convergence=self._convergence_data )
+                                 convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
 
                 return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
-                             convergence=self._convergence_data )
+                             convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
             elif msg['action'] == 'stop':
                 self.__stop( )
                 return dict( result='stopped', update=dict( ) )
@@ -511,7 +521,7 @@ class InteractiveClean:
                                                                  ids=self._ids['clean'] ),
                                                       code=self._js['initialize'] )
 
-        self._fig['convergence'] = figure( tooltips=[ ("x","$x"), ("y","$y"), ("value", "@image")],
+        self._fig['convergence'] = figure( tooltips=[ ("x","$x"), ("y","$y"), ("value", "(@peakRes, @modelFlux)")],
                                            output_backend="webgl", plot_height=180, plot_width=800,
                                            tools=[ ],
                                            title='Convergence', x_axis_label='Iteration', y_axis_label='Peak Residual' )
@@ -562,6 +572,7 @@ class InteractiveClean:
                                         icon=SVGIcon(icon_name='help', size=1.4) )
 
 
+        self._control['nmajor'] = TextInput( title='nmajor', value="%s" % self._params['nmajor'], width=90 )
         self._control['niter'] = TextInput( title='niter', value="%s" % self._params['niter'], width=90 )
         self._control['cycleniter'] = TextInput( title="cycleniter", value="%s" % self._params['cycleniter'], width=90 )
         self._control['threshold'] = TextInput( title="threshold", value="%s" % self._params['threshold'], width=90 )
@@ -581,6 +592,7 @@ class InteractiveClean:
                                                  img_src=self._fig['image-source'],
                                                  #spec_src=self._image_spectra,
                                                  niter=self._control['niter'], cycleniter=self._control['cycleniter'],
+                                                 nmajor=self._control['nmajor'],
                                                  threshold=self._control['threshold'], cyclefactor=self._control['cycle_factor'],
                                                  convergence_src=self._convergence_source, convergence_id=self._convergence_id,
                                                  convergence_fig=self._fig['convergence'],
@@ -589,7 +601,7 @@ class InteractiveClean:
                                                  image_fig=self._fig['image'],
                                                  spectra_fig=self._fig['spectra'],
                                                  stopstatus=self._status['stopcode'],
-                                                 #stat_src=self._stats_source
+                                                 cube_obj = self._cube.js_obj( )
                                                 ),
                                       code=self._js['update-converge'] + self._js['clean-refresh'] + self._js['clean-disable'] +
                                            self._js['clean-enable'] + self._js['clean-status-update'] +
@@ -601,7 +613,7 @@ class InteractiveClean:
                                                   btns['stop'].button_type = "warning"
                                                   ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                   { action: 'finish',
-                                                                    value: { niter: niter.value, cycleniter: cycleniter.value,
+                                                                    value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                              threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                              mask: img_src.masks( ),
                                                                              breadcrumbs: img_src.breadcrumbs( ) } },
@@ -616,7 +628,7 @@ class InteractiveClean:
                                                   // (or even if 'XXX.origin.' is public)...
                                                   ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                   { action: 'next',
-                                                                    value: { niter: niter.value, cycleniter: cycleniter.value,
+                                                                    value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                              threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                              mask: img_src.masks( ),
                                                                              breadcrumbs: img_src.breadcrumbs( ) } },
@@ -654,25 +666,27 @@ class InteractiveClean:
                                                             self._js['update-converge'] + self._js['slider-update'] ) )
         
         # Generates the HTML for the controls layout:
-        # niter cycleniter cycle_factor threshold  -----------
-        #                  slider                  -  image  -
-        # goto               continue finish stop  -----------
-        # status: stopcode                  stats         help
+        # nmajor niter cycleniter cycle_factor threshold  -----------
+        #                  slider                         -  image  -    help
+        # goto    continue finish stop                    -----------    help
+        # status: stopcode                                               help
+        # stats                                                          help
         # spectra
         # convergence
         # log
         self._fig['layout'] = column(
                                   row(
                                       column(
-                                          row( self._control['niter'],
-                                               self._control['cycleniter'],
-                                               self._control['cycle_factor'],
-                                               self._control['threshold'] ),
-                                          self._fig['slider'],
-                                          row( self._control['goto'],
+                                          row( self._control['nmajor'],
+                                               self._control['niter'],
+                                               Spacer(width=10, sizing_mode='scale_height'),
                                                self._control['clean']['continue'],
-                                               self._control['clean']['finish'],
+                                               self._control['clean']['finish'] ),
+                                          row( self._control['cycleniter'],
+                                               self._control['cycle_factor'],
+                                               self._control['threshold'],
                                                self._control['clean']['stop'] ),
+                                          row( self._fig['slider'], self._control['goto'] ),
                                           row ( Div( text="<div><b>status:</b></div>" ), self._status['stopcode'] ),
                                           self._cube.statistics( sizing_mode = "stretch_height" ),
                                       ),
@@ -697,6 +711,46 @@ class InteractiveClean:
         self._control['help'].js_on_click( CustomJS( args=dict( help=self._cube.help( ) ),
                                                      code='''if ( help.visible == true ) help.visible = false
                                                              else help.visible = true''' ) )
+
+        self._control['goto'].js_on_change( 'value', CustomJS( args=dict( img=self._cube.js_obj( ),
+                                                                          slider=self._fig['slider'],
+                                                                          status=self._status['stopcode'] ),
+                                                               code='''let values = cb_obj.value.split(/[ ,]+/).map((v,) => parseInt(v))
+                                                                       if ( values.length > 2 ) {
+                                                                         status._error_set = true
+                                                                         status.text = '<div>enter at most two indexes</div>'
+                                                                       } else if ( values.filter((x) => x < 0 || isNaN(x)).length > 0 ) {
+                                                                         status._error_set = true
+                                                                         status.text = '<div>invalid channel entered</div>'
+                                                                       } else {
+                                                                         if ( status._error_set ) {
+                                                                           status._error_set = false
+                                                                           status.text = '<div/>'
+                                                                         }
+                                                                         if ( values.length == 1 ) {
+                                                                           if ( values[0] >= 0 && values[0] < img.num_chans[1] ) {
+                                                                             status.text= `<div>moving to channel ${values[0]}</div>`
+                                                                             slider.value = values[0]
+                                                                           } else {
+                                                                             status._error_set = true
+                                                                             status.text = `<div>channel ${values[0]} out of range</div>`
+                                                                           }
+                                                                         } else if ( values.length == 2 ) {
+                                                                           if ( values[0] < 0 || values[0] >= img.num_chans[1] ) {
+                                                                             status._error_set = true
+                                                                             status.text = `<div>channel ${values[0]} out of range</div>`
+                                                                           } else {
+                                                                             if ( values[1] < 0 || values[1] >= img.num_chans[0] ) {
+                                                                               status._error_set = true
+                                                                               status.text = `<div>stokes ${values[1]} out of range</div>`
+                                                                             } else {
+                                                                               status.text= `<div>moving to channel ${values[0]}/${values[1]}</div>`
+                                                                               slider.value = values[0]
+                                                                               img.channel( values[0], values[1] )
+                                                                             }
+                                                                           }
+                                                                         }
+                                                                       }''' ) )
         self._cube.connect( )
         show(self._fig['layout'])
 
